from .error import DeviceStateNotAvailable
from typing import Any, TYPE_CHECKING, Union
from blebox_uniapi.jfollow import follow

if TYPE_CHECKING:
    from .box import Box


class Feature:
    _device_class: str

    def __init__(self, product: "Box", alias: str, methods: dict):
        self._product = product
        self._alias = alias
        self._methods = methods

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Feature"]:
        # note: by default single config entry yields single feature instance but certain feature
        # domains (e.g. lights) may handle this differently depending on their `extended_state`
        return [cls(product, *args) for args in box_type_config]

    @property
    def unique_id(self) -> str:
        return f"BleBox-{self._product.type}-{self._product.unique_id}-{self._alias}"

    async def async_update(self) -> None:
        await self._product.async_update_data()

    @property
    def full_name(self) -> str:
        product = self._product
        return f"{product.name} ({product.type}#{self._alias})"

    @property
    def device_class(self) -> str:
        return self._device_class

    @property
    def product(self) -> "Box":
        return self._product

    @property
    def alias(self):
        return self._alias

    # TODO: (cleanup) move to product/box ?
    def raw_value(self, name: str) -> Any:
        product = self._product
        # TODO: better exception?
        if product.last_data is None:
            # TODO: coverage
            raise DeviceStateNotAvailable  # pragma: no cover
        methods = self._methods
        if method := methods.get(name):
            return follow(product.last_data, method)
        return None

    async def async_api_command(self, *args: Any, **kwargs: Any) -> None:
        await self._product.async_api_command(*args, **kwargs)

    @staticmethod
    def resolve_access_method_paths(
        methods: dict[str, Union[str, callable]], id_val: str = None
    ) -> dict[str, str]:
        """Return dict with resolved callable used as data path."""
        new = dict()
        if not isinstance(methods, dict):
            raise TypeError(
                f"Parameter methods should be dict, instead of {type(methods)}."
            )
        for key, value in methods.items():
            if callable(value):
                new[key] = value(id_val)
            else:
                new[key] = value
        return new



"""Tests for `blebox_uniapi` package."""

import pytest
import logging
import aiohttp

from unittest.mock import patch, Mock, AsyncMock

from blebox_uniapi.session import ApiHost as Session
from blebox_uniapi import error


@pytest.fixture
def mocked_client():
    with patch("aiohttp.ClientSession", spec_set=True, autospec=True) as mocked_session:
        yield mocked_session.return_value


@pytest.fixture
def logger():
    return Mock(spec_set=logging.Logger).return_value


@pytest.fixture
def client():
    return Mock(spec_set=aiohttp.ClientSession)


def valid_response():
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = 200
    response.content_type = "application/json"
    response.text = AsyncMock(return_value="foobar")
    response.json = AsyncMock(return_value=123)
    return response


def timeout_error(connection, timeout):
    raise aiohttp.ServerTimeoutError


def client_error(connection, timeout):
    raise aiohttp.ClientError("client err")


def os_error(connection, timeout):
    raise aiohttp.ClientOSError("os error")


def bad_http_response(spec_set=aiohttp.ClientResponse):
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = 400
    return response


async def test_session_api_get(logger, client):
    client.get = AsyncMock(return_value=valid_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)

    result = await api_session.async_api_get("/api/foo")

    client.get.assert_called_once_with("http://127.0.0.4:88/api/foo", timeout=2)

    assert result == 123


async def test_session_default_client_created(mocked_client, logger):
    mocked_client.get = AsyncMock(return_value=valid_response())
    api_session = Session("127.0.0.4", "88", 2, None, None, logger)

    result = await api_session.async_api_get("/api/foo")

    mocked_client.get.assert_called_once_with("http://127.0.0.4:88/api/foo", timeout=2)
    assert result == 123


async def test_session_default_timeout_used(mocked_client, logger):
    mocked_client.get = AsyncMock(return_value=valid_response())
    api_session = Session("127.0.0.4", "88", None, None, None, logger)

    await api_session.async_api_get("/api/foo")
    expected_timeout = aiohttp.ClientTimeout(
        total=None, connect=None, sock_read=5, sock_connect=5
    )

    mocked_client.get.assert_called_once_with(
        "http://127.0.0.4:88/api/foo", timeout=expected_timeout
    )


async def test_session_api_get_timeout(logger, client):
    client.get = AsyncMock(side_effect=timeout_error)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)

    with pytest.raises(error.TimeoutError):
        await api_session.async_api_get("/api/foo")


async def test_session_api_post_timeout(logger, client):
    def post_timeout_error(connection, **kwargs):
        timeout_error(connection, timeout=kwargs.get("timeout"))

    client.post = AsyncMock(side_effect=post_timeout_error)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)

    with pytest.raises(error.TimeoutError):
        await api_session.async_api_post("/api/foo", {})


async def test_session_api_get_client_error(logger, client):
    client.get = AsyncMock(side_effect=client_error)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(
        error.ClientError,
        match=r"API request http://127\.0\.0\.4:88/api/foo failed: client err",
    ):
        await api_session.async_api_get("/api/foo")


async def test_session_always_show_address_details(logger, client):
    client.get = AsyncMock(side_effect=os_error)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(
        error.ConnectionError, match=r"Failed to connect to 127\.0\.0\.4:88: os error"
    ):
        await api_session.async_api_get("/api/foo")


async def test_session_api_post_client_error(logger, client):
    def post_client_error(connection, **kwargs):
        client_error(connection, timeout=kwargs.get("timeout"))

    client.post = AsyncMock(side_effect=post_client_error)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.ClientError):
        await api_session.async_api_post("/api/foo", {})


async def test_session_api_get_http_error(logger, client):
    client.get = AsyncMock(return_value=bad_http_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.HttpError):
        await api_session.async_api_get("/api/foo")


async def test_session_api_post_http_error(logger, client):
    client.post = AsyncMock(return_value=bad_http_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.HttpError):
        await api_session.async_api_post("/api/foo", {})


async def test_session_provides_a_logger(logger, client):
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    api_session.logger.debug("foobar")
    logger.debug.assert_called_once_with("foobar")


def ota_accepted_response(status=202):
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = status
    response.content_type = None
    return response


def unauthorized_response():
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = 401
    return response


def non_json_response():
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = 200
    response.content_type = None
    return response


async def test_session_api_get_ota_accepts_202_without_body(logger, client):
    client.get = AsyncMock(return_value=ota_accepted_response(status=202))
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    result = await api_session.async_api_get_ota("/api/ota/check")
    assert result is None


async def test_session_api_get_ota_accepts_204_no_content(logger, client):
    client.get = AsyncMock(return_value=ota_accepted_response(status=204))
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    result = await api_session.async_api_get_ota("/api/ota/check")
    assert result is None


async def test_session_api_get_ota_rejects_400(logger, client):
    client.get = AsyncMock(return_value=bad_http_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.HttpError):
        await api_session.async_api_get_ota("/api/ota/check")


async def test_session_api_get_unauthorized(logger, client):
    client.get = AsyncMock(return_value=unauthorized_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.UnauthorizedRequest):
        await api_session.async_api_get("/api/device/state")


async def test_session_api_get_ota_unauthorized(logger, client):
    client.get = AsyncMock(return_value=unauthorized_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.UnauthorizedRequest):
        await api_session.async_api_get_ota("/api/ota/check")


async def test_session_api_get_non_json_returns_none(logger, client):
    client.get = AsyncMock(return_value=non_json_response())
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    result = await api_session.async_api_get("/api/device/state")
    assert result is None


async def test_session_api_get_unicode_decode_error_raises_connection_error(logger, client):
    response = Mock(spec_set=aiohttp.ClientResponse)
    response.status = 200
    response.content_type = "application/json"
    response.json = AsyncMock(side_effect=UnicodeDecodeError("utf-8", b"", 0, 1, "invalid"))
    client.get = AsyncMock(return_value=response)
    api_session = Session("127.0.0.4", "88", 2, client, None, logger)
    with pytest.raises(error.ConnectionError, match="Invalid response encoding"):
        await api_session.async_api_get("/api/device/state")



from random import choice

from blebox_uniapi.box_types import (
    BOX_TYPE_CONF,
    get_conf,
    get_conf_set,
    get_latest_api_level,
    get_latest_conf,
)


class TestBoxTypesOrder:
    def test_conf_order(self):
        for device_type, conf_set in BOX_TYPE_CONF.items():
            api_levels = list(conf_set.keys())
            sorted_api_levels = sorted(api_levels)

            assert (
                api_levels == sorted_api_levels
            ), f"Entries for '{device_type}' are not ordered by API level."


class TestBoxTypes:
    box_types = tuple(BOX_TYPE_CONF.keys())
    simple_conf_set = {5: {"tag": "first_entry"}, 10: {"tag": "second_entry"}}

    async def test_get_conf_set_valid(self):
        conf_set = get_conf_set(choice(self.box_types))

        assert isinstance(conf_set, dict)
        assert conf_set != {}

    async def test_get_conf_set_invalid(self):
        conf_set = get_conf_set("nonexistent_type")

        assert isinstance(conf_set, dict)
        assert conf_set == {}

    async def test_get_conf_valid(self):
        """Test choosing functionality of get_conf function on exemplary conf_set."""
        for api_level, tag_value in (
            (5, "first_entry"),  # a marginal example
            (7, "first_entry"),  # 'in between' example
            (10, "second_entry"),  # a marginal example
            (17, "second_entry"),  # future example
        ):
            conf = get_conf(api_level, self.simple_conf_set)

            assert isinstance(conf, dict)
            assert conf["tag"] == tag_value

    async def test_get_conf_invalid(self):
        conf = get_conf(3, self.simple_conf_set)  # not supported example

        assert isinstance(conf, dict)
        assert conf == {}

    async def test_get_latest_conf_valid(self):
        conf = get_latest_conf(choice(self.box_types))

        assert isinstance(conf, dict)
        assert conf != {}

    async def test_get_latest_conf_invalid(self):
        conf = get_latest_conf("nonexistent_type")

        assert isinstance(conf, dict)
        assert conf == {}

    async def test_get_latest_api_level_valid(self):
        api_level = get_latest_api_level(choice(self.box_types))

        assert isinstance(api_level, int)
        assert api_level

    async def test_get_latest_api_level_invalid(self):
        api_level = get_latest_api_level("nonexistent_type")

        assert isinstance(api_level, int)
        assert not api_level



"""Blebox switch tests."""

from blebox_uniapi.box_types import get_latest_api_level

from .conftest import CommonEntity, DefaultBoxTest, future_date

DEVICE_CLASS_SWITCH = "switch"


class BleBoxSwitchEntity(CommonEntity):
    @property
    def device_class(self):
        types = {"relay": DEVICE_CLASS_SWITCH}
        return types[self._feature.device_class]

    @property
    def is_on(self):
        return self._feature.is_on

    async def async_turn_on(self, **kwargs):
        return await self._feature.async_turn_on()

    async def async_turn_off(self, **kwargs):
        return await self._feature.async_turn_off()


class TestSwitchBox0(DefaultBoxTest):
    """Tests for BleBox switchBox."""

    DEVCLASS = "switches"
    ENTITY_CLASS = BleBoxSwitchEntity

    DEV_INFO_PATH = "api/relay/state"

    DEVICE_INFO = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": "20180604",
        }
    }
    DEVICE_INFO_FUTURE = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": future_date(),
        }
    }
    DEVICE_INFO_LATEST = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": get_latest_api_level("switchBox"),
        }
    }
    DEVICE_INFO_UNSUPPORTED = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": 20180603,
        }
    }
    DEVICE_INFO_UNSPECIFIED_API = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
        }
    }

    STATE_OFF = [{"relay": 0, "state": 0, "stateAfterRestart": 0}]
    STATE_ON = [{"relay": 0, "state": 1, "stateAfterRestart": 0}]
    STATE_DEFAULT = STATE_OFF

    async def test_init(self, aioclient_mock):
        """Test switch default state."""
        await self.allow_get_info(aioclient_mock)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My switchBox (switchBox#0.relay)"
        assert entity.unique_id == "BleBox-switchBox-1afe34e750b8-0.relay"
        assert entity.device_class == DEVICE_CLASS_SWITCH
        assert entity.is_on is None

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.device_info["name"] == "My switchBox"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "switchBox"
        assert entity.device_info["sw_version"] == "0.247"

    async def test_update_when_off(self, aioclient_mock):
        """Test switch updating when off."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

    async def test_update_when_on(self, aioclient_mock):
        """Test switch updating when on."""
        entity = await self.updated(aioclient_mock, self.STATE_ON)
        assert entity.is_on is True

    async def test_on(self, aioclient_mock):
        """Test turning switch on."""
        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        self.allow_get(aioclient_mock, "/s/1", self.STATE_ON)
        await entity.async_turn_on()
        assert entity.is_on is True

    async def test_off(self, aioclient_mock):
        """Test turning switch off."""
        entity = await self.updated(aioclient_mock, self.STATE_ON)
        self.allow_get(aioclient_mock, "/s/0", self.STATE_OFF)
        await entity.async_turn_off()
        assert entity.is_on is False


class TestSwitchBox(DefaultBoxTest):
    """Tests for BleBox switchBox."""

    DEVCLASS = "switches"
    ENTITY_CLASS = BleBoxSwitchEntity

    DEV_INFO_PATH = "api/relay/extended/state"
    DEVICE_INFO = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": "20190808",
        }
    }

    def patch_version(apiLevel):
        """Generate a patch for a JSON state fixture."""
        return f"""
        {{ "device": {{ "apiLevel": {apiLevel} }} }}
        """

    DEVICE_INFO_FUTURE = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": future_date(),
        }
    }
    DEVICE_INFO_LATEST = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            "apiLevel": get_latest_api_level("switchBox"),
        }
    }
    DEVICE_INFO_UNSUPPORTED = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
            # since below it 20180808 it switches to switchBox0
            "apiLevel": 20180604 - 1,
        }
    }

    DEVICE_INFO_UNSPECIFIED_API = {
        "device": {
            "deviceName": "My switchBox",
            "type": "switchBox",
            "fv": "0.247",
            "hv": "0.2",
            "id": "1afe34e750b8",
            "ip": "192.168.1.239",
        }
    }

    STATE_OFF = {"relays": [{"relay": 0, "state": 0, "stateAfterRestart": 0}]}
    STATE_ON = {"relays": [{"relay": 0, "state": 1, "stateAfterRestart": 0}]}
    STATE_DEFAULT = STATE_OFF

    async def test_init(self, aioclient_mock):
        """Test switch default state."""

        await self.allow_get_info(aioclient_mock)

        entity = (await self.async_entities(aioclient_mock))[0]

        assert entity.name == "My switchBox (switchBox#0.relay)"
        assert entity.unique_id == "BleBox-switchBox-1afe34e750b8-0.relay"

        assert entity.device_class == DEVICE_CLASS_SWITCH

        assert entity.is_on is None

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My switchBox"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "switchBox"
        assert entity.device_info["sw_version"] == "0.247"

    async def test_update_when_off(self, aioclient_mock):
        """Test switch updating when off."""

        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        assert entity.is_on is False

    async def test_update_when_on(self, aioclient_mock):
        """Test switch updating when on."""

        entity = await self.updated(aioclient_mock, self.STATE_ON)
        assert entity.is_on is True

    async def test_on(self, aioclient_mock):
        """Test turning switch on."""

        entity = await self.updated(aioclient_mock, self.STATE_OFF)
        self.allow_get(aioclient_mock, "/s/1", self.STATE_ON)
        await entity.async_turn_on()
        assert entity.is_on is True

    async def test_off(self, aioclient_mock):
        """Test turning switch off."""

        entity = await self.updated(aioclient_mock, self.STATE_ON)
        self.allow_get(aioclient_mock, "/s/0", self.STATE_OFF)
        await entity.async_turn_off()
        assert entity.is_on is False


class TestSwitchBoxD(DefaultBoxTest):
    """Tests for BleBox switchBoxD."""

    DEVCLASS = "switches"
    ENTITY_CLASS = BleBoxSwitchEntity
    DEV_INFO_PATH = "state/extended"
    DEVICE_INFO = {
        "device": {
            "deviceName": "My switchBoxD",
            "type": "switchBoxD",
            "fv": "0.200",
            "hv": "0.7",
            "id": "1afe34e750b8",
            "apiLevel": "20200831",
        }
    }
    DEVICE_INFO_FUTURE = {
        "device": {
            "deviceName": "My switchBoxD",
            "type": "switchBoxD",
            "fv": "0.200",
            "hv": "0.7",
            "id": "1afe34e750b8",
            "apiLevel": future_date(),
        }
    }
    DEVICE_INFO_LATEST = {
        "device": {
            "deviceName": "My switchBoxD",
            "type": "switchBoxD",
            "fv": "0.200",
            "hv": "0.7",
            "id": "1afe34e750b8",
            "apiLevel": get_latest_api_level("switchBoxD"),
        }
    }

    DEVICE_INFO_UNSUPPORTED = {
        "device": {
            "deviceName": "My switchBoxD",
            "type": "switchBoxD",
            "fv": "0.200",
            "hv": "0.7",
            "id": "1afe34e750b8",
            "apiLevel": 20190807,
        }
    }

    DEVICE_INFO_UNSPECIFIED_API = {
        "device": {
            "deviceName": "My switchBoxD",
            "type": "switchBoxD",
            "fv": "0.200",
            "hv": "0.7",
            "id": "1afe34e750b8",
        }
    }
    STATE_BOTH_OFF = {
        "relays": [
            {"relay": 0, "state": 0, "stateAfterRestart": 0, "name": "output 1"},
            {"relay": 1, "state": 0, "stateAfterRestart": 0, "name": "output 2"},
        ]
    }
    STATE_FIRST_ON = {
        "relays": [
            {"relay": 0, "state": 1, "stateAfterRestart": 0, "name": "output 1"},
            {"relay": 1, "state": 0, "stateAfterRestart": 0, "name": "output 2"},
        ]
    }
    STATE_SECOND_ON = {
        "relays": [
            {"relay": 0, "state": 0, "stateAfterRestart": 0, "name": "output 1"},
            {"relay": 1, "state": 1, "stateAfterRestart": 0, "name": "output 2"},
        ]
    }
    STATE_BOTH_ON = {
        "relays": [
            {"relay": 0, "state": 1, "stateAfterRestart": 0, "name": "output 1"},
            {"relay": 1, "state": 1, "stateAfterRestart": 0, "name": "output 2"},
        ]
    }
    STATE_DEFAULT = STATE_BOTH_OFF
    DEVICE_EXTENDED_INFO = STATE_BOTH_OFF
    DEVICE_EXTENDED_INFO_PATH = "/state/extended"

    async def test_init(self, aioclient_mock):
        """Test switch default state."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        entity = entities[0]
        assert entity.unique_id == "BleBox-switchBoxD-1afe34e750b8-0.relay_0"
        assert entity.device_class == DEVICE_CLASS_SWITCH
        assert entity._feature.index == 0
        assert entity._feature.name == "output 1"

        entity = entities[1]
        assert entity.unique_id == "BleBox-switchBoxD-1afe34e750b8-0.relay_1"
        assert entity.device_class == DEVICE_CLASS_SWITCH
        assert entity._feature.index == 1
        assert entity._feature.name == "output 2"

    async def test_device_info(self, aioclient_mock):
        await self.allow_get_info(aioclient_mock, self.DEVICE_INFO)
        entity = (await self.async_entities(aioclient_mock))[0]
        assert entity.device_info["name"] == "My switchBoxD"
        assert entity.device_info["mac"] == "1afe34e750b8"
        assert entity.device_info["manufacturer"] == "BleBox"
        assert entity.device_info["model"] == "switchBoxD"
        assert entity.device_info["sw_version"] == "0.200"

    async def test_update_when_off(self, aioclient_mock):
        """Test switch updating when off."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get_state(aioclient_mock, self.STATE_BOTH_OFF)

        # updating any one is fine
        await entities[0].async_update()

        assert entities[0].is_on is False
        assert entities[1].is_on is False

    async def test_update_when_second_off(self, aioclient_mock):
        """Test switch updating when off."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get_state(aioclient_mock, self.STATE_FIRST_ON)

        # updating any one is fine
        await entities[0].async_update()

        assert entities[0].is_on is True
        assert entities[1].is_on is False

    async def test_first_on(self, aioclient_mock):
        """Test turning switch on."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get(aioclient_mock, "/s/0/1", self.STATE_FIRST_ON)
        await entities[0].async_turn_on()
        assert entities[0].is_on is True
        assert entities[1].is_on is False

    async def test_second_on(self, aioclient_mock):
        """Test turning switch on."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get(aioclient_mock, "/s/1/1", self.STATE_SECOND_ON)
        await entities[1].async_turn_on()
        assert entities[0].is_on is False
        assert entities[1].is_on is True

    async def test_first_off(self, aioclient_mock):
        """Test turning switch on."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get_state(aioclient_mock, self.STATE_BOTH_ON)
        await entities[0].async_update()

        self.allow_get(aioclient_mock, "/s/0/0", self.STATE_SECOND_ON)
        await entities[0].async_turn_off()
        assert entities[0].is_on is False
        assert entities[1].is_on is True

    async def test_second_off(self, aioclient_mock):
        """Test turning switch on."""
        await self.allow_get_info(aioclient_mock)
        entities = await self.async_entities(aioclient_mock)

        self.allow_get_state(aioclient_mock, self.STATE_BOTH_ON)
        await entities[0].async_update()

        self.allow_get(aioclient_mock, "/s/1/0", self.STATE_FIRST_ON)
        await entities[1].async_turn_off()
        assert entities[0].is_on is True
        assert entities[1].is_on is False



from __future__ import annotations

import asyncio
import time

from typing import Optional, Any, Dict

from .box_types import _DEFAULT_API_LEVEL, get_conf, get_conf_set
from .button import Button
from .climate import Climate
from .cover import Cover
from .light import Light
from .sensor import SensorFactory
from .binary_sensor import BinarySensor
from .session import ApiHost
from .switch import Switch
from .update import Update

from .error import (
    UnsupportedBoxResponse,
    UnsupportedBoxVersion,
    BadFieldExceedsMax,
    BadFieldLessThanMin,
    BadFieldMissing,
    BadFieldNotANumber,
    BadFieldNotAString,
    BadFieldNotRGBW,
    HttpError,
)

DEFAULT_PORT = 80


class Box:
    _name: str
    _data_path: str
    _last_real_update: Optional[float]
    _last_data: Optional[Any]

    api_session: ApiHost
    info: dict
    config: dict
    extended_state: Optional[Dict[Any, Any]]

    def __init__(
        self,
        api_session,
        info,
        config,
        extended_state,
    ) -> None:
        self._last_data = None
        self._last_real_update = None
        self._sem = asyncio.BoundedSemaphore()
        self._session = api_session

        address = f"{api_session.host}:{api_session.port}"

        location = f"Device at {address}"

        # NOTE: get ID first for better error messages
        try:
            unique_id = info["id"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(info, f"{location} has no id") from ex
        location = f"Device:{unique_id} at {address}"

        try:
            type = info["type"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(info, f"{location} has no type") from ex

        try:
            product = info["product"]
        except KeyError:
            product = type

        # TODO: make wLightBox API support multiple products
        # in 2020 wLightBoxS API has been deprecated and it started using wLightBox API
        # current codebase needs a refactor to support multiple product sharing one API
        # as a temporary workaround we are using 'alias' type wLightBoxS2
        if type == "wLightBox" and product == "wLightBoxS":
            type = "wLightBoxS"

        location = f"{product}:{unique_id} at {address}"

        try:
            name = info["deviceName"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(info, f"{location} has no name") from ex
        location = f"'{name}' ({product}:{unique_id} at {address})"

        try:
            firmware_version = info["fv"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(
                info, f"{location} has no firmware version"
            ) from ex
        location = f"'{name}' ({product}:{unique_id}/{firmware_version} at {address})"

        try:
            hardware_version = info["hv"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(
                info, f"{location} has no hardware version"
            ) from ex

        available_firmware_version = info.get("availableFv")

        level = int(info.get("apiLevel", _DEFAULT_API_LEVEL))

        self._data_path = config["api_path"]
        self._type = type
        self._unique_id = unique_id
        self._name = name
        self._address = address
        self._firmware_version = firmware_version
        self._hardware_version = hardware_version
        self._available_firmware_version = available_firmware_version
        self._api_version = level
        self._model = config.get("model", type)
        self._api = config.get("api", {})
        self._features = self.create_features(config, info, extended_state)
        self._config = config
        self._update_last_data(extended_state)

    def create_features(
        self, config: dict, info: dict, extended_state: Optional[dict]
    ) -> dict:
        features = {}
        for field, klass in {
            "covers": Cover,
            "sensors": SensorFactory,
            "binary_sensors": BinarySensor,
            "lights": Light,
            "climates": Climate,
            "switches": Switch,
            "buttons": Button,
        }.items():
            try:
                if box_type_config := config.get(field):
                    features[field] = klass.many_from_config(
                        self,
                        box_type_config=box_type_config,
                        extended_state=extended_state,
                    )
            except KeyError:
                raise UnsupportedBoxResponse("Failed to initialize:", info)
        features["updates"] = [Update(self, "update", {})]
        return features

    @classmethod
    async def async_from_host(cls, api_host: ApiHost) -> Box:
        try:
            path = "/api/device/state"
            data = await api_host.async_api_get(path)
        except HttpError:
            path = "/info"
            data = await api_host.async_api_get(path)
        if data is None:
            raise UnsupportedBoxResponse("Device returned non-JSON response")
        info = data.get("device", data)
        extended_state = None

        config = cls._match_device_config(info)
        if extended_state_path := config.get("extended_state_path"):
            try:
                extended_state = await api_host.async_api_get(extended_state_path)
            except (HttpError, KeyError):
                extended_state = None

        return cls(api_host, info, config, extended_state)

    @classmethod
    def _match_device_config(cls, info: dict) -> dict:
        try:
            device_type = info["type"]
        except KeyError as ex:
            raise UnsupportedBoxResponse(info, "Device info has no type key") from ex
        try:
            product = info["product"]
        except KeyError:
            product = device_type
        if device_type == "wLightBox" and product == "wLightBoxS":
            device_type = "wLightBoxS"
        level = int(info.get("apiLevel", _DEFAULT_API_LEVEL))
        config_set = get_conf_set(device_type)
        if not config_set:
            raise UnsupportedBoxResponse(f"{device_type} is not a supported type")
        config = get_conf(level, config_set)
        if not config:
            raise UnsupportedBoxVersion(
                f"{device_type} has unsupported version ({level})."
            )

        return config

    @property
    def name(self) -> str:
        return self._name

    @property
    def address(self) -> str:
        return self._address

    @property
    def last_data(self) -> Optional[Dict[Any, Any]]:
        return self._last_data

    # Used in full_name, track down and refactor.
    @property
    def type(self) -> str:
        return self._type

    @property
    def unique_id(self) -> Any:
        return self._unique_id

    @property
    def firmware_version(self) -> Any:
        return self._firmware_version

    @property
    def hardware_version(self) -> Any:
        return self._hardware_version

    @property
    def available_firmware_version(self) -> Any:
        return self._available_firmware_version

    @property
    def api_version(self) -> int:
        return self._api_version

    @property
    def features(self) -> dict:
        return self._features

    @property
    def brand(self) -> str:
        return "BleBox"

    @property
    def model(self) -> Any:
        return self._model

    async def async_update_data(self) -> None:
        await self._async_api(True, "GET", self._data_path)

    def _update_last_data(self, new_data: Optional[dict]) -> None:
        # Note: on certain more complex devices that inlcude multiple features
        # like switches and sensors (e.g. SwitchboxD) it may happen that activating
        # single feature would result only in partial update of the self._last_data.
        # We can know that by comparing keys of both states.
        #
        # Notable example of device exhibiting this behavior is SwitchboxD (20200831)
        # It has two relays (switches) and power measurement capabilities (sensor).
        # Toggling the relay does return state of all relays, but does not return
        # sensor information (partial state). Accepting the new state as-is would
        # break the update of sensory information.
        #
        # Note that SwitchboxD is just an example. It is possible that APIs of other
        # box types also exhibit this kind of behavior.
        if (
            isinstance(self._last_data, dict)
            and isinstance(new_data, dict)
            and self._last_data.keys() != new_data.keys()
        ):
            # ... In such a case we need to merge both states instead of overwriting
            # the old one as-is.
            #
            # The only risk is that if certain features are somehow coupled inside
            # the device, we will have inconsistent information about the device.
            # However, this should be eventually consistent as new updates arrive.
            # In the end, it is better to have an inconsistent state, rather than
            # have broken one (e.g. missing keys) that results in broken features.
            #
            # Refs:
            # - https://github.com/blebox/blebox_uniapi/pull/152
            # - https://github.com/blebox/blebox_uniapi/issues/137
            new_data = {**self._last_data, **new_data}

        self._last_data = new_data
        for feature_set in self._features.values():
            for feature in feature_set:
                feature.after_update()

    async def async_api_command(self, command: str, value: Any = None) -> None:
        method, *args = self._api[command](value)
        self._last_real_update = None  # force update
        return await self._async_api(False, method, *args)

    def expect_int(
        self, field: str, raw_value: int, maximum: int = -1, minimum: int = 0
    ) -> int:
        return self.check_int(raw_value, field, maximum, minimum)

    def expect_hex_str(
        self, field: str, raw_value: int, maximum: int = -1, minimum: int = 0
    ) -> int:
        return self.check_hex_str(raw_value, field, maximum, minimum)

    def expect_rgbw(self, field: str, raw_value: int) -> int:
        return self.check_rgbw(raw_value, field)

    def check_int_range(
        self, value: int, field: str, max_value: int, min_value: int
    ) -> int:
        if max_value >= min_value:
            if value > max_value:
                raise BadFieldExceedsMax(self.name, field, value, max_value)
            if value < min_value:
                raise BadFieldLessThanMin(self.name, field, value, min_value)
        return value

    def check_int(self, value: int, field: str, max_value: int, min_value: int) -> int:
        if value is None:
            raise BadFieldMissing(self.name, field)

        if not isinstance(value, int):
            raise BadFieldNotANumber(self.name, field, value)

        return self.check_int_range(value, field, max_value, min_value)

    def check_hex_str(
        self, value: int, field: str, max_value: int, min_value: int
    ) -> int:
        if value is None:
            raise BadFieldMissing(self.name, field)

        if not isinstance(value, str):
            raise BadFieldNotAString(self.name, field, value)

        if (
            self.check_int_range(int(value, 16), field, max_value, min_value)
            is not None
        ):
            return value

    def check_rgbw(self, value: int, field: str) -> int:
        if value is None:
            raise BadFieldMissing(self.name, field)

        if not isinstance(value, str):
            raise BadFieldNotAString(self.name, field, value)

        if len(value) > 10 or len(value) % 2 != 0:
            raise BadFieldNotRGBW(self.name, field, value)
        return value

    def _has_recent_data(self) -> bool:
        last = self._last_real_update
        return (time.time() - 2) <= last if last is not None else False

    async def _async_api(
        self,
        is_update: bool,
        method: Any,
        path: str,
        post_data: dict = None,
    ) -> None:
        if method not in ("GET", "POST"):
            raise NotImplementedError(method)  # pragma: no cover

        if is_update:
            if self._has_recent_data():
                return

        async with self._sem:
            if is_update:
                if self._has_recent_data():
                    return
            if method == "GET":
                response = await self._session.async_api_get(path)
            else:
                response = await self._session.async_api_post(path, post_data)
            self._update_last_data(response)
            self._last_real_update = time.time()

    async def async_ota_check(self) -> None:
        await self._session.async_api_get_ota("/api/ota/check")
        for _ in range(3):
            await asyncio.sleep(1)
            response = await self._session.async_api_get("/info")
            if response is None:
                raise UnsupportedBoxResponse("Device returned non-JSON response")
            response = response.get("device", response)
            if response.get("availableFv") is not None:
                self._firmware_version = response.get("fv", self._firmware_version)
                self._available_firmware_version = response["availableFv"]
                return

    async def async_ota_update(self) -> None:
        await self._session.async_api_get("/api/ota/update")



from unittest.mock import Mock
from blebox_uniapi.box_types import BOX_TYPE_CONF
import pytest

from blebox_uniapi.light import Light
from blebox_uniapi.box import Box


@pytest.fixture
def product():
    return Mock(spec=Box)


@pytest.fixture
def dimmer_box(product):
    product.type = "dimmerBox"
    many = Light.many_from_config(
        product, BOX_TYPE_CONF["dimmerBox"][20170829]["lights"], extended_state={}
    )
    assert len(many) == 1
    return many[0]


@pytest.fixture
def w_light_box_rgbww(product):
    product.type = "wLightBox"
    extended_state = {
        "rgbw": {
            "colorMode": 7,
            "effectID": 0,
            "desiredColor": "fcfffcff00",
            "currentColor": "fcfffcff00",
            "lastOnColor": "fcfffcff00",
            "durationsMs": {"colorFade": 1000, "effectFade": 1000, "effectStep": 1000},
            "favColors": {"0": "ff", "1": "00", "2": "c0", "3": "40", "4": "00"},
            "effectsNames": {"0": "NONE", "1": "FADE", "2": "Stroboskop", "3": "BELL"},
        }
    }
    many = Light.many_from_config(
        product,
        BOX_TYPE_CONF["wLightBox"][20200229]["lights"],
        extended_state=extended_state,
    )
    return many[0]


@pytest.mark.parametrize("input_on", range(1, 255))
async def test_dimmer_box_async_on_with_int(dimmer_box: Light, product: Mock, input_on):
    await dimmer_box.async_on(input_on)
    product.async_api_command.assert_called_with("set", input_on)


async def test_dimmer_box_async_on_zero(dimmer_box: Light, product: Mock):
    with pytest.raises(ValueError):
        await dimmer_box.async_on(0)


async def test_dimmer_box_async_on_with_hex(dimmer_box: Light, product: Mock):
    await dimmer_box.async_on("ff")
    product.async_api_command.assert_called_with("set", 255)


@pytest.mark.parametrize("input_on", [255, "ff"])
async def test_dimmer_box_async_on(dimmer_box: Light, product: Mock, input_on):
    await dimmer_box.async_on(input_on)
    product.async_api_command.assert_called_with("set", 255)


async def test_dimmer_box_async_off(dimmer_box: Light, product: Mock):
    await dimmer_box.async_off()
    product.async_api_command.assert_called_with("set", 0)


@pytest.mark.parametrize(
    "io_params",
    [("ff", [255]), ("ffff", [255, 255]), ("ff14cdfe6a", [255, 20, 205, 254, 106])],
)
def test_dimmer_box_rgb_hex_to_rgb_list(dimmer_box: Light, io_params):
    assert dimmer_box.rgb_hex_to_rgb_list(io_params[0]) == io_params[1]


@pytest.mark.parametrize(
    "io_params",
    [
        (["ff"], [255]),
        (["ff", "ff"], [255, 255]),
        (["ff", "14", "cd", "fe", "6a"], [255, 20, 205, 254, 106]),
    ],
)
def test_dimmer_box_rgb_list_to_rgb_hex_list(dimmer_box: Light, io_params):
    assert dimmer_box.rgb_list_to_rgb_hex_list(io_params[1]) == io_params[0]


@pytest.mark.parametrize(
    "io_params",
    [
        ([255], [255]),
        ([255, 255], [255, 255]),
        ([120, 20, 205, 96, 106], [149, 25, 255, 119, 132]),
    ],
)
def test_dimmer_box_normalise_elements_of_rgb(dimmer_box: Light, io_params):
    assert dimmer_box.normalise_elements_of_rgb(io_params[0]) == io_params[1]


def test_dimmer_box_normalise_elements_of_rgb_invalid_values(dimmer_box: Light):
    with pytest.raises(ValueError):
        dimmer_box.normalise_elements_of_rgb([-10])

    with pytest.raises(ValueError):
        dimmer_box.normalise_elements_of_rgb([256])

    with pytest.raises(ValueError):
        dimmer_box.normalise_elements_of_rgb([-1, 0])


@pytest.mark.parametrize(
    "io_params", [([145, 135, 90], 145), ([145, 135, 240], 240), ([255], 255)]
)
def test_dimmer_box_evaluate_brightness_from_rgb(dimmer_box: Light, io_params):
    assert dimmer_box.evaluate_brightness_from_rgb(io_params[0]) == io_params[1]


def test_dimmer_box_apply_brightness_zero(dimmer_box: Light):
    assert dimmer_box.apply_brightness(10, 0) == [0]


def test_dimmer_box_evaluate_brightness_from_rgb_out_of_range(dimmer_box: Light):
    with pytest.raises(ValueError):
        dimmer_box.evaluate_brightness_from_rgb([257, 135, 90])

    with pytest.raises(ValueError):
        dimmer_box.evaluate_brightness_from_rgb([145, 135, -1])

    with pytest.raises(ValueError):
        dimmer_box.evaluate_brightness_from_rgb([145, -10, 900])


def test_light_sensible_on_value_last_is_zero(w_light_box_rgbww: Light):
    assert len(w_light_box_rgbww.effect_list) == 4
    assert w_light_box_rgbww.effect == "NONE"



from .cover import Gate, GateBox, GateBoxB, Shutter
from typing import Union, Any

# default api level for all products that don't have one
_DEFAULT_API_LEVEL = 20151206


def get_conf_set(product_type: str) -> dict:
    """Get all configurations for provided product type."""
    conf_set = BOX_TYPE_CONF.get(product_type, {})
    return conf_set


def get_conf(api_level: Union[int, str], conf_set: dict) -> dict:
    """Get configuration from conf_set for provided api_level."""
    for min_api_level in sorted(conf_set, reverse=True):
        if api_level >= min_api_level:
            return conf_set[min_api_level]

    return {}


def get_latest_conf(product_type: str) -> dict:
    """Get latest configuration for provided product type."""
    conf_set = get_conf_set(product_type)
    if conf_set:
        latest_min_api_level = sorted(conf_set, reverse=True)[0]
        return conf_set[latest_min_api_level]

    return conf_set


def get_latest_api_level(product_type: str) -> Union[dict, int]:
    """Get latest supported api_level for provided product type."""
    conf_set = get_conf_set(product_type)
    if conf_set:
        return sorted(conf_set, reverse=True)[0]
    return 0


# Configuration for all box types

BOX_TYPE_CONF: dict[str, dict[int, dict[str, Any]]] = {
    # tvLiftBox; in comments api level config description
    "tvLiftBox": {  # apiType to match devices apiType
        20200518: {  # apiLevel to match integration level
            "api_path": "/api/device/state",  # path to devices state
            "extended_state_path": "/state/extended",  # path to devices extended state
            "api": {
                "set": lambda command: ("GET", f"/s/c/{command}")
            },  # dictionary with interaction methods
            "buttons": [
                "tvLift",
                {"lift": ""},
            ],  # key used to set platform, list elements used in cls init, e.g. [<alias>, {"path": "state_value"}]
        }
    },
    # airSensor
    "airSensor": {
        20180403: {
            "api_path": "/api/air/state",
            "sensors": [
                [
                    "0.air",
                    {
                        "pm1.value": "air.sensors[?type == 'pm1']|[0]|value",
                        "pm1.state": "air.sensors[?type == 'pm1']|[0]|state",
                        "pm2_5.value": "air.sensors[?type == 'pm2.5']|[0]|value",
                        "pm2_5.state": "air.sensors[?type == 'pm2.5']|[0]|state",
                        "pm10.value": "air.sensors[?type == 'pm10']|[0]|value",
                        "pm10.state": "air.sensors[?type == 'pm10']|[0]|state",
                    },
                ]
            ],
        }
    },
    # dimmerBox
    "dimmerBox": {
        # default_API
        20151206: {
            "api_path": "/api/dimmer/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/dimmer/set",
                    '{"dimmer":{"desiredBrightness": ' + str(x) + "}}",
                ),
            },
            "lights": [["brightness", {"desired": "dimmer.desiredBrightness"}]],
        },
        20170829: {
            "api_path": "/api/dimmer/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/dimmer/set",
                    '{"dimmer":{"desiredBrightness": ' + str(x) + "}}",
                ),
            },
            "lights": [["brightness", {"desired": "dimmer.desiredBrightness"}]],
        },
    },
    # gateBox
    "gateBox": {
        # default_API
        20151206: {
            "api_path": "/api/gate/state",
            "api": {
                "primary": lambda x=None: ("GET", "/s/p", None),
                "secondary": lambda x=None: ("GET", "/s/s", None),
            },
            "covers": [
                [
                    "position",
                    {
                        "position": "currentPos",
                        "desired": "desiredPos",
                        "extraButtonType": "extraButtonType",
                    },
                    "gatebox",
                    GateBox,
                ]
            ],
        },
        20200831: {
            "api_path": "/state/extended",
            "extended_state_path": "/state/extended",
            "api": {
                "primary": lambda x=None: ("GET", "/s/p", None),
                "secondary": lambda x=None: ("GET", "/s/s", None),
            },
            "covers": [
                [
                    "position",
                    {
                        "position": "gate.currentPos",
                        "gate_type": "gate.gateType",
                    },
                    "gatebox",
                    GateBoxB,
                ]
            ],
        },
    },
    # gateController
    "gateController": {
        20180604: {
            "api_path": "/api/gatecontroller/state",
            "extended_state_path": "/api/gatecontroller/extended/state",
            "api": {
                "open": lambda x=None: ("GET", "/s/o", None),
                "close": lambda x=None: ("GET", "/s/c", None),
                "position": lambda x: ("GET", "/s/p/" + str(x), None),
                "stop": lambda x=None: ("GET", "/s/s", None),
                # "walk": lambda x=None: ("GET", "/s/w", None),
                # "next": lambda x=None: ("GET", "/s/n", None),
            },
            "covers": [
                [
                    "position",
                    {
                        "desired": "gateController.desiredPos.positions|[0]",
                        # "current": "gateController/currentPos/positions/[0]",
                        "state": "gateController.state",
                    },
                    "gate",
                    Gate,
                ]
            ],
        }
    },
    # thermoBox
    "thermoBox": {
        20200229: {
            "api_path": "/state/extended",
            "extended_state_path": "/state/extended",
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
                "set": lambda x=None: ("GET", "/s/t/" + str(x), None),
            },
            "climates": [
                [
                    "thermostat",
                    {
                        "desired": "thermo.desiredTemp",
                        "minimum": "thermo.minimumTemp",
                        "maximum": "thermo.maximumTemp",
                        "temperature": lambda x: f"sensors[?id == `{x}`]|[0]|value",
                        "state": "thermo.state",
                        "mode": "thermo.mode",
                        "safetySensorId": "thermo.safetyTempSensor.sensorId",
                        "operatingState": "thermo.operatingState",
                    },
                ]
            ],
            "sensors": [
                [
                    "temperature",
                    {
                        "temperature": lambda x: f"sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
        }
    },
    # saunaBox
    "saunaBox": {
        20180604: {
            # TODO: read extended state only once on startup
            "api_path": "/api/heat/extended/state",
            "extended_state_path": "/api/heat/extended/state",
            # TODO: use an api map (map to semver)? Or constraints?
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
                "set": lambda x=None: ("GET", "/s/t/" + str(x), None),
            },
            "climates": [
                [
                    "thermostat",
                    {
                        "desired": "heat.desiredTemp",
                        "minimum": "heat.minimumTemp",
                        "maximum": "heat.maximumTemp",
                        "temperature": "heat.sensors[?id == `0`]|[0]|value",
                        "state": "heat.state",
                    },
                ]
            ],
        }
    },
    # shutterBox
    "shutterBox": {
        20180604: {
            "api_path": "/api/shutter/state",
            "extended_state_path": "/api/shutter/extended/state",
            "api": {
                "open": lambda x=None: ("GET", "/s/u", None),
                "close": lambda x=None: ("GET", "/s/d", None),
                "position": lambda x: ("GET", "/s/p/" + str(x), None),
                "stop": lambda x=None: ("GET", "/s/s", None),
                "tilt": lambda x=None: ("GET", "/s/t/" + str(x), None),
            },
            "covers": [
                [
                    "position",
                    {
                        "desired": "shutter.desiredPos.position",
                        # "current": "shutter/currentPos/position",
                        "tilt": "shutter.desiredPos.tilt",
                        "state": "shutter.state",
                    },
                    "shutter",
                    Shutter,
                ]
            ],
        }
    },
    # switchBox
    "switchBox": {
        20180604: {
            "model": "switchBox",
            "api_path": "/api/relay/state",
            "extended_state_path": "/api/relay/extended/state",
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
            },
            "switches": [["0.relay", {"state": "[?relay==`0`]|[0]|state"}, "relay"]],
            # note: does not support power measurement
        },
        20190808: {
            "api_path": "/api/relay/extended/state",
            "extended_state_path": "/api/relay/extended/state",
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                ]
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20200229: {
            "extended_state_path": "/state/extended",
            "api_path": "/state/extended",
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                ]
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20200831: {
            "extended_state_path": "/state/extended",
            "api_path": "/state/extended",
            "api": {
                "on": lambda x=None: ("GET", "/s/1", None),
                "off": lambda x=None: ("GET", "/s/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                ]
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20220114: {
            "api_path": "/state/extended",
            "extended_state_path": "/state/extended",
            "api": {
                # note: old control api (i.e. /s/0, /s/1, /s/2) still supported but
                # now deprecated. switchBox has now API consistent with switchBoxD
                "on": lambda x=None: ("GET", f"/s/{x}/1", None),
                "off": lambda x=None: ("GET", f"/s/{x}/0", None),
            },
            "switches": [
                [
                    "relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                ]
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value ",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20220505: {
            "api_path": "/state/extended",
            "extended_state_path": "/state/extended",
            "api": {
                # note: old control api (i.e. /s/0, /s/1, /s/2) still supported but
                # now deprecated. switchBox has now API consistent with switchBoxD
                "on": lambda x=None: ("GET", f"/s/{x}/1", None),
                "off": lambda x=None: ("GET", f"/s/{x}/0", None),
            },
            "switches": [
                [
                    "relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                ]
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value ",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "voltage": lambda x: "sensors[?type == 'voltage']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
    },
    # switchBoxD
    "switchBoxD": {
        20190808: {
            "extended_state_path": "/api/relay/extended/state",
            "api_path": "/api/relay/extended/state",
            "api": {
                "on": lambda x: ("GET", f"/s/{int(x)}/1", None),
                "off": lambda x=None: ("GET", f"/s/{int(x)}/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    0,
                ],
                [
                    "1.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    1,
                ],
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value ",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20200229: {
            "extended_state_path": "/state/extended",
            "api_path": "/state/extended",
            "api": {
                "on": lambda x: ("GET", f"/s/{int(x)}/1", None),
                "off": lambda x=None: ("GET", f"/s/{int(x)}/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    0,
                ],
                [
                    "1.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    1,
                ],
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value ",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
        20200831: {
            "extended_state_path": "/state/extended",
            "api_path": "/state/extended",
            "api": {
                "on": lambda x: ("GET", f"/s/{int(x)}/1", None),
                "off": lambda x=None: ("GET", f"/s/{int(x)}/0", None),
            },
            "switches": [
                [
                    "0.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    0,
                ],
                [
                    "1.relay",
                    {"state": lambda x: f"relays[?relay==`{x}`]|[0]|state"},
                    "relay",
                    1,
                ],
            ],
            "sensors": [
                [
                    "switchBox.energy",
                    {
                        # note: switchbox/switchboxD sensors are currently not indexed (singletons)
                        "powerConsumption": lambda x: "powerMeasuring.powerConsumption[0]|value ",
                        "activePower": lambda x: "sensors[?type == 'activePower']|[0]|value",
                        "periodS": "powerMeasuring.powerConsumption[0]|periodS",
                        "measurement_enabled": "powerMeasuring.enabled",
                    },
                ]
            ],
        },
    },
    # tempSensor
    "tempSensor": {
        20180604: {
            "api_path": "/api/tempsensor/state",
            "sensors": [
                [
                    "0.temperature",
                    {
                        "temperature": "tempSensor.sensors[?id == `0`]|[0]|value",
                        "trend": "tempSensor.sensors[?id == `0`]|[0]|trend",
                        "state": "tempSensor.sensors[?id == `0`]|[0]|state",
                        "elapsed": "tempSensor.sensors[?id == `0`]|[0]|elapsedTimeS",
                    },
                ]
            ],
        }
    },
    # wLightBox
    "wLightBox": {
        # default_API
        20151206: {
            "api_path": "/api/device/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/rgbw/set",
                    f'{{"rgbw":{{"desiredColor": "{str(x)}"}}}}',
                ),
            },
            "lights": [
                [
                    "color",
                    {
                        "desired": "rgbw.desiredColor",
                        "last_color": "rgbw.lastOnColor",
                        "currentEffect": "rgbw.effectID",
                        "colorMode": "rgbw.colorMode",
                    },
                ]
            ],
        },
        20190808: {
            "api_path": "/api/rgbw/state",
            "extended_state_path": "/api/rgbw/extended/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/rgbw/set",
                    f'{{"rgbw":{{"desiredColor": "{str(x)}"}}}}',
                ),
                "effect": lambda x: ("GET", f"/s/x/{x}", None),
            },
            "lights": [
                [
                    "color",
                    {
                        "desired": "rgbw.desiredColor",
                        "last_color": "rgbw.lastOnColor",
                        "currentEffect": "rgbw.effectID",
                        "colorMode": "rgbw.colorMode",
                    },
                ]
            ],
        },
        20200229: {
            "api_path": "/api/rgbw/state",
            "extended_state_path": "/api/rgbw/extended/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/rgbw/set",
                    f'{{"rgbw": {{"desiredColor": "{x}"}}}}',
                ),
                "effect": lambda x: ("GET", f"/s/x/{x}"),
            },
            "lights": [
                [
                    "color",
                    {
                        "desired": "rgbw.desiredColor",
                        "last_color": "rgbw.lastOnColor",
                        "currentEffect": "rgbw.effectID",
                        "colorMode": "rgbw.colorMode",
                    },
                ],
            ],
        },
    },
    # wLightBoxS
    "wLightBoxS": {
        # default_API
        20151206: {
            "api_path": "/api/device/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/light/set",
                    f'{{"light": {{"desiredColor": "{x}"}}}}',
                ),
            },
            "lights": [
                [
                    "brightness",
                    {
                        "desired": "light.desiredColor",
                    },
                ]
            ],
        },
        20180718: {
            "api_path": "/api/light/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/light/set",
                    f'{{"light": {{"desiredColor": "{x}"}}}}',
                ),
                "effect": lambda x: ("GET", f"/s/x/{x}"),
            },
            "lights": [
                [
                    "brightness",
                    {
                        "desired": "light.desiredColor",
                    },
                ]
            ],
        },
        20200229: {
            "api_path": "/api/rgbw/state",
            "extended_state_path": "/api/rgbw/extended/state",
            "api": {
                "set": lambda x: (
                    "POST",
                    "/api/rgbw/set",
                    f'{{"rgbw": {{"desiredColor": "{x}"}}}}',
                ),
                "effect": lambda x: ("GET", f"/s/x/{x}"),
            },
            "lights": [
                [
                    "brightness",
                    {
                        "desired": "rgbw.desiredColor",
                        "colorMode": "rgbw.colorMode",
                        "currentEffect": "rgbw.effectID",
                        "last_color": "rgbw.lastOnColor",
                    },
                ]
            ],
        },
    },
    # multiSensor
    "multiSensor": {
        20200831: {
            "api_path": "/state",
            "extended_state_path": "/state/extended",
            "sensors": [
                [
                    "multiSensor",
                    {
                        "temperature": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "wind": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
            "binary_sensors": [
                [
                    "multiSensor",
                    {
                        "rain": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "flood": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
        },
        20210413: {
            "api_path": "/state",
            "extended_state_path": "/state/extended",
            "sensors": [
                [
                    "multiSensor",
                    {
                        "temperature": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "wind": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "humidity": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
            "binary_sensors": [
                [
                    "multiSensor",
                    {
                        "rain": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "flood": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
        },
        20220114: {
            "api_path": "/state",
            "extended_state_path": "/state/extended",
            "sensors": [
                [
                    "multiSensor",
                    {
                        "illuminance": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "temperature": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "wind": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'wind']|[0]|value",
                        "humidity": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
            "binary_sensors": [
                [
                    "multiSensor",
                    {
                        "rain": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                        "flood": lambda x: f"multiSensor.sensors[?id == `{x}`]|[0]|value",
                    },
                ]
            ],
        },
        20230606: {
            "api_path": "/state",
            "extended_state_path": "/state/extended",
            "sensors": [
                [
                    "multiSensor",
                    {
                        "frequency": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'frequency']|[0]|value",
                        "current": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'current']|[0]|value",
                        "voltage": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'voltage']|[0]|value",
                        "apparentPower": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'apparentPower']|[0]|value",
                        "activePower": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'activePower']|[0]|value",
                        "reactivePower": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'reactivePower']|[0]|value",
                        "reverseActiveEnergy": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'reverseActiveEnergy']|[0]|value",
                        "forwardActiveEnergy": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'forwardActiveEnergy']|[0]|value",
                        "forwardReactiveEnergy": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'forwardReactiveEnergy']|[0]|value",
                        "reverseReactiveEnergy": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'reverseReactiveEnergy']|[0]|value",
                        "illuminance": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'illuminance']|[0]|value",
                        "temperature": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'temperature']|[0]|value",
                        "wind": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'wind']|[0]|value",
                        "humidity": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'humidity']|[0]|value",
                        "openStatus": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'openStatus']|[0]|value",
                        "co2": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'co2']|[0]|value",
                        "co2Definition": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'co2Definition']|[0]|value",
                    },
                ]
            ],
            "binary_sensors": [
                [
                    "multiSensor",
                    {
                        "rain": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'rain']|[0]|value",
                        "flood": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'flood']|[0]|value",
                        "openStatus": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'openStatus']|[0]|value",
                        "binary": lambda x: f"multiSensor.sensors[?id == `{x}`]|[?type == 'binary']|[0]|value",
                    },
                ]
            ],
        },
    },
}



from .feature import Feature
from typing import TYPE_CHECKING, Optional, Any, Union

if TYPE_CHECKING:
    from .box import Box


class Switch(Feature):
    _is_on: Optional[bool]

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        dev_class: str,
        unit_id: Union[str, int, None] = 0,
        name: Optional[str] = None,
    ):
        methods = self.resolve_access_method_paths(methods, str(unit_id))
        super().__init__(product, alias, methods)
        self._device_class = dev_class
        self._unit_id = unit_id
        self._name = name

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Switch"]:
        """
        :param product: Object hosting device with specific feature.
        :param box_type_config: Default configuration providing following data
        [
            [feature_alias, {method_name: method_path_or_method_path_callable}, relay_type, unit_id]
        ]
        :param extended_state: Object hosting extended state recieved from device
        :return: List of class objects instances
        """
        if extended_state:
            relay_list = list()
            alias, methods, relay_type, *rest = box_type_config[0]
            for relay in extended_state.get("relays", []):
                relay_id = relay.get("relay")
                relay_list.append(
                    cls(
                        product,
                        alias + "_" + str(relay_id),
                        methods,
                        relay_type,
                        relay_id,
                        name=relay.get("name"),
                    )
                )

            return relay_list
        else:
            return [cls(product, *args) for args in box_type_config]

    def after_update(self) -> None:
        self._is_on = self._read_is_on()

    def _read_is_on(self) -> Optional[bool]:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value("state")
            if raw is not None:  # no reading
                alias = self._alias
                return 1 == product.expect_int(alias, raw, 1, 0)
        return None

    @property
    def is_on(self) -> Optional[bool]:
        return self._is_on

    @property
    def _unit_args(self) -> list:
        unit = self._unit_id
        return [] if unit is None else [unit]

    @property
    def index(self) -> Optional[int]:
        return self._unit_id

    @property
    def name(self) -> Optional[str]:
        return self._name

    async def async_turn_on(self, **kwargs: Any) -> None:
        await self.async_api_command("on", *self._unit_args)

    async def async_turn_off(self, **kwargs: Any) -> None:
        await self.async_api_command("off", *self._unit_args)



from typing import Any, Optional, Union

import aiohttp
import asyncio
import logging

from . import error

DEFAULT_TIMEOUT = aiohttp.ClientTimeout(total=None, sock_connect=5, sock_read=5)
DEFAULT_PORT = 80

logger = logging.getLogger(__name__)


class ApiHost:
    def __init__(
        self,
        host: str,
        port: int,
        timeout: int,
        session: Any,
        loop: Any,
        logger: logging.Logger = logger,
        **auth,
    ):
        self._host = host
        self._port = port
        self._username = auth.get("username")
        self._password = auth.get("password")
        # TODO: handle empty logger?
        self._logger = logger

        self._timeout = timeout if timeout else DEFAULT_TIMEOUT

        self._session = session

        auth = None

        if any(data is not None for data in [self._username, self._password]):
            auth = aiohttp.BasicAuth(login=self._username, password=self._password)

        if not self._session:
            self._session = aiohttp.ClientSession(loop=loop, timeout=timeout, auth=auth)

        # TODO: remove?
        self._loop = loop

    async def async_request(
        self,
        path: str,
        async_method: Any,
        data: Union[dict, str, None] = None,
        allow_ota_check_response: bool = False,
    ) -> Optional[dict]:
        # TODO: check timeout
        client_timeout = self._timeout
        url = self.api_path(path)
        try:
            if data is not None:
                response = await async_method(url, timeout=client_timeout, data=data)
            else:
                response = await async_method(url, timeout=client_timeout)

            accepted_statuses = (200, 202, 204) if allow_ota_check_response else (200,)
            if response.status not in accepted_statuses:
                if response.status == 401:
                    raise error.UnauthorizedRequest(
                        f"Request to {url} failed with HTTP {response.status}, UNAUTHORISED"
                    )
                raise error.HttpError(
                    f"Request to {url} failed with HTTP {response.status}"
                )

            if response.content_type != "application/json":
                return None

            return await response.json()

        except asyncio.TimeoutError as ex:
            raise error.TimeoutError(
                f"Failed to connect to {self.host}:{self.port} within {client_timeout}s: ({ex})"
            ) from ex

        except aiohttp.ClientConnectionError as ex:
            raise error.ConnectionError(
                f"Failed to connect to {self.host}:{self.port}: {ex}"
            ) from ex

        except aiohttp.ClientError as ex:
            raise error.ClientError(f"API request {url} failed: {ex}") from ex

        except UnicodeDecodeError as ex:
            raise error.ConnectionError(
                f"Invalid response encoding from {url}: {ex}"
            ) from ex

    async def async_api_get(self, path: str) -> Optional[dict]:
        try:
            return await self.async_request(path, self._session.get)
        except Exception as ex:
            logger.error(f"EXCEPTION DURING API CALL: {ex}")
            raise ex

    async def async_api_get_ota(self, path: str) -> Optional[dict]:
        try:
            return await self.async_request(
                path, self._session.get, allow_ota_check_response=True
            )
        except Exception as ex:
            logger.error(f"EXCEPTION DURING API CALL: {ex}")
            raise ex

    async def async_api_post(
        self, path: str, data: Union[dict, str, None]
    ) -> Optional[dict]:
        return await self.async_request(path, self._session.post, data)

    def api_path(self, path: str) -> str:
        host = self._host
        port = self._port

        # TODO: url lib
        return f"http://{host}:{port}/{path[1:]}"

    @property
    def logger(self) -> Any:
        return self._logger

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port



import datetime
import numbers
from functools import partial

from .feature import Feature
from typing import TYPE_CHECKING, Union, Optional

if TYPE_CHECKING:
    from .box import Box


class SensorFactory:
    device_constructors: dict[str, type] = {}

    @classmethod
    def register(cls, sensor_type: str, **kwargs):
        if sensor_type in cls.device_constructors:
            raise RuntimeError("Can't register same sensor type twice")

        def decorator(registrable: type):
            constructor = registrable
            if kwargs:
                constructor = partial(registrable, sensor_type=sensor_type, **kwargs)

            cls.device_constructors[sensor_type] = constructor
            # note: returning unmodified, so we can register registrable
            # multiple times under different names and with different kwargs
            return registrable

        return decorator

    @staticmethod
    def _sensor_states(extended_state: dict):
        """Read potential sensor states from extended state dictionary"""
        # note: probably we should iterate extended state in future if there
        # are other api flavours other than multiSensor that provide sensors
        states = extended_state.get("multiSensor", {}).get("sensors", [])
        # note: but for now we are only able to support non-multisensor devices
        # that provide sensor data in extended data payload root
        states.extend(extended_state.get("sensors", []))
        # note: power measuring feature predates multiSensor API, so we need a small
        # shim to adapt older shape of power measuring schema to the new sensor API
        if "powerMeasuring" in extended_state:
            power_states = extended_state["powerMeasuring"].get("powerConsumption", [])
            # note: be careful of names as this has been historically named differently
            # in home-assistant
            states.extend({"type": "powerConsumption", **s} for s in power_states)
        return states

    @classmethod
    def many_from_config(cls, product, box_type_config, extended_state):
        if extended_state:
            object_list = []
            # note: first item was historically an alias, but it has been since
            # abandoned. We still keep it in the box config.
            _, methods = box_type_config[0]

            for sensor in cls._sensor_states(extended_state):
                device_class = sensor.get("type")
                sensor_id = sensor.get("id")

                alias = device_class
                if sensor_id is not None:
                    alias = f"{device_class}_{sensor_id}"

                if constructor := cls.device_constructors.get(device_class):
                    # note: methods for sensor readings are provided as template
                    # functions (lambdas) in the box config. We need to "materialize"
                    # them to make sure they are properly indexed by sensor ID
                    materialized_methods = {
                        **methods,
                        device_class: methods[device_class](sensor_id),
                    }

                    feature = constructor(
                        product=product,
                        alias=alias,
                        methods=materialized_methods,
                        sensor_id=sensor_id,
                        name=sensor.get("name"),
                    )
                    object_list.append(feature)

            return object_list

        # legacy handling of some old device API that do not provide extended state
        alias, methods = box_type_config[0]
        if alias.endswith("air"):
            method_list = [method for method in methods if "value" in method]
            return [
                AirQuality(product=product, alias=method.split(".")[0], methods=methods)
                for method in method_list
            ]
        if alias.endswith("temperature"):
            return [Temperature(product=product, alias=alias, methods=methods)]
        else:
            return []


class BaseSensor(Feature):
    _unit: str
    _device_class: str
    _native_value: Union[float, int, str]
    _sensor_type: Optional[str]
    _sensor_id: Optional[int]

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_type: str = None,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        self._sensor_type = sensor_type
        self._sensor_id = sensor_id
        self._name = name
        super().__init__(product, alias, methods)

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def device_class(self) -> str:
        return self._device_class

    @property
    def native_value(self):
        return self._native_value

    @property
    def sensor_id(self):
        return self._sensor_id

    @property
    def probe_id(self):
        return self.sensor_id

    @property
    def index(self) -> Optional[int]:
        return self._sensor_id

    @property
    def name(self) -> Optional[str]:
        return self._name

    @classmethod
    def many_from_config(cls, product, box_type_config, extended_state):
        raise NotImplementedError("Please use SensorFactory")

    def __str__(self):
        return f"<{self.__class__.__name__} sensor_type={self._sensor_type}, alias={self._alias}>"


@SensorFactory.register("frequency", unit="Hz", scale=1_000)
@SensorFactory.register("current", unit="mA")
@SensorFactory.register("voltage", unit="V", scale=10)
@SensorFactory.register("apparentPower", unit="va")
@SensorFactory.register("reactivePower", unit="var")
@SensorFactory.register("activePower", unit="W")
@SensorFactory.register("reverseReactiveEnergy", unit="kvarh", scale=1_000)
@SensorFactory.register("forwardReactiveEnergy", unit="kvarh", scale=1_000)
@SensorFactory.register("reverseActiveEnergy", unit="kWh", scale=1_000)
@SensorFactory.register("forwardActiveEnergy", unit="kWh", scale=1_000)
@SensorFactory.register("illuminance", unit="lx", scale=100)
@SensorFactory.register("humidity", unit="percentage", scale=100)
@SensorFactory.register("wind", unit="m/s", scale=10)
@SensorFactory.register("openStatus", unit="")
@SensorFactory.register("co2", unit="ppm")
@SensorFactory.register("co2Definition", unit="")
class GenericSensor(BaseSensor):
    def __init__(
        # base sensor params
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int],
        name: Optional[str] = None,
        *,
        # generalization params
        sensor_type: str,
        unit: str,
        scale: float = 1,
        precision: Optional[int] = None,
    ):
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)
        self._unit = unit
        self._scale = scale
        self._precision = precision
        # note: this seems redundant but there is at least one sensor type that
        # has different mapping in home assistant (wind/wind_speed). Should be
        # fixed in upstream first.
        self._device_class = sensor_type
        self._sensor_type = sensor_type

    def after_update(self):
        product = self._product
        if product.last_data is None:
            return

        raw = self.raw_value(self._device_class)
        if not isinstance(raw, numbers.Number):
            raw = float("nan")

        native = raw / self._scale
        if self._precision:
            native = round(native, self._precision)

        self._native_value = native


@SensorFactory.register("powerConsumption", unit="kWh")
class PowerConsumption(GenericSensor):
    # note: almost the same as typical generic sensor but also provides extra property
    # to read last reset value
    @property
    def last_reset(self):
        return datetime.datetime.now() - datetime.timedelta(
            seconds=self._read_period_of_measurement()
        )

    def _read_period_of_measurement(self) -> int:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value("periodS")
            if raw is not None:
                alias = self._alias
                return product.expect_int(alias, raw, 3600, 0)
        return 0


@SensorFactory.register("temperature")
class Temperature(BaseSensor):
    _current: Union[float, int, None]

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[int] = None,
        name: Optional[str] = None,
    ):
        super().__init__(product, alias, methods, sensor_id=sensor_id, name=name)
        self._unit = "celsius"
        self._device_class = "temperature"

    @property
    def current(self) -> Union[float, int, None]:
        return self._current

    def _read_temperature(self, field: str) -> Union[float, int, None]:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value(field)
            if raw is not None:
                alias = self._alias
                return round(product.expect_int(alias, raw, 12500, -5500) / 100.0, 1)
        return None

    def after_update(self) -> None:
        self._current = self._read_temperature("temperature")
        self._native_value = self._read_temperature("temperature")


@SensorFactory.register("airSensor")
class AirQuality(BaseSensor):
    _pm: Optional[int]

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        sensor_id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__(product, alias, methods, sensor_id, name=name)
        self._unit = "concentration_of_mp"
        self._device_class = alias

    def _pm_value(self, name: str) -> Optional[int]:
        product = self._product
        if product.last_data is not None:
            raw = self.raw_value(name)
            if raw is not None:
                alias = self._alias
                return product.expect_int(alias, raw, 3000, 0)
        return None

    def after_update(self) -> None:
        self._native_value = self._pm_value(f"{self.device_class}.value")



from typing import Any, Union

import jmespath


def follow(data: Union[dict, list], path: str) -> Any:
    if data is None:
        raise RuntimeError(f"bad argument: data {data}")  # pragma: no cover

    expression = jmespath.compile(path)
    return expression.search(data)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["aiohttp>=3", "jmespath>1.0.0"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest>=3",
    "pytest-asyncio>=0.10.0",
    "deepmerge",
]

setup(
    version="2.5.5",
    author="BleBox",
    author_email="opensource@blebox.eu",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Python API for accessing BleBox smart home devices",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=False,
    keywords="blebox_uniapi",
    name="blebox_uniapi",
    packages=find_packages(include=["blebox_uniapi", "blebox_uniapi.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/blebox/blebox_uniapi",
    zip_safe=False,
)



from typing import Optional

from .feature import Feature


class Update(Feature):
    @property
    def installed_version(self) -> Optional[str]:
        return self._product.firmware_version

    @property
    def latest_version(self) -> Optional[str]:
        return self._product.available_firmware_version

    async def async_update(self) -> None:
        # OTA state is stored as Box attributes, not in last_data. async_update_data() won't trigger the OTA check.
        await self._product.async_ota_check()

    async def async_install(self) -> None:
        await self._product.async_ota_update()

    def after_update(self) -> None:
        pass



from enum import IntEnum, auto

from .error import MisconfiguredDevice
from .feature import Feature
from typing import TYPE_CHECKING, Any, Optional, Type, TypeVar

if TYPE_CHECKING:
    from .box import Box


class BleboxCoverState(IntEnum):
    """BleboxCoverState defines possible states of cover devices.

    Note that enumeration of states is partially shared between
    different types of devices (shutterBox, gateController) but
    not all states are possible for every type. For details of
    states refer to blebox official API documentation.
    """

    MOVING_DOWN = 0
    MOVING_UP = 1
    MANUALLY_STOPPED = 2
    LOWER_LIMIT_REACHED = 3
    UPPER_LIMIT_REACHED = 4
    OVERLOAD = 5
    MOTOR_FAILURE = 6
    UNUSED = 7
    SAFETY_STOP = 8


class ShutterBoxControlType(IntEnum):
    """ShutterBoxControlType defines shuterBox command semantics"""

    SEGMENTED_SHUTTER = 1
    NO_CALIBRATION = 2
    TILT_SHUTTER_90 = 3
    WINDOW_OPENER = 4
    MATERIAL_SHUTTER = 5
    AWNING = 6
    SCREEN = 7
    CURTAIN = 8
    ROOF_SEGMENTED_SHUTTER = 9
    SERVOMOTOR = 10
    CURTAIN_ONE_SIDED = 11
    CURTAIN_TWO_SIDED = 12
    TILT_ONLY = 13
    PERGOLA_ROOF = 14
    PERGOLA_ROOF_TILT_ONLY = 15
    TILT_SHUTTER_180 = 16
    TILT_SHUTTER_WITHOUT_POSITIONING = 17


class GateBoxControlType(IntEnum):
    """GateBoxControlType defines gateBox command semantics known as `openCloseMode`.

    Control type affects mainly [o]pen, [c]lose, and [n]ext commands which
    are wrappers around [p]rimary and [s]econdary outputs. The only exception
    is OPEN_CLOSE (2) control type that also means that the gateBox lacks
    stop action because typical stop output is wired to [c]lose/[s]econdary
    command.
    """

    STEP_BY_STEP = 0
    ONLY_OPEN = 1
    OPEN_CLOSE = 2


class GateBoxGateType(IntEnum):
    """GateBoxGateType defines possible gate/cover types reported by gateBox"""

    SLIDING_DOOR = 0
    GARAGE_DOOR = 1
    OVER_DOOR = 2
    DOOR = 3


class UnifiedCoverType(IntEnum):
    """UnifiedCoverType defines single "cover type" concept shared between different
    devices.

    Some device types have concept of control type/mode that affects how device
    operates and how it is being used (e.g. control type in shutterBox/gateControler),
    but others have these two concepts separated (e.g. open mode vs. gate type in
    gateBox). This enum provides unified concept of controlled cover type that
    can be infered from internal device information end exposed to library user.

    """

    AWNING = auto()
    BLIND = auto()
    CURTAIN = auto()
    DAMPER = auto()
    DOOR = auto()
    GARAGE = auto()
    GATE = auto()
    SHADE = auto()
    SHUTTER = auto()
    WINDOW = auto()


class Gate:
    _control_type: Optional[int]

    def __init__(self, control_type: int, is_calibrated: bool = False):
        self._control_type = control_type

    def read_state(self, alias: str, raw_value: Any, product: "Box") -> int:
        raw = raw_value("state")
        return product.expect_int(alias, raw, max(BleboxCoverState).value, 0)

    def read_desired(self, alias: str, raw_value: Any, product: "Box") -> Optional[int]:
        raw = raw_value("desired")
        min_position = self.min_position
        return product.expect_int(alias, raw, 100, min_position)

    def read_tilt(self, alias: str, raw_value: Any, product: "Box") -> int:
        raw = raw_value("tilt")
        min_position = self.min_position
        return product.expect_int(alias, raw, 100, min_position)

    def read_cover_type(
        self, alias: str, raw_value: Any, product: "Box"
    ) -> UnifiedCoverType:
        return UnifiedCoverType.GATE

    @property
    def min_position(self) -> int:
        return 0

    @property
    def is_position_inverted(self) -> bool:
        # shutterBox/gateController: % closed (0=open, 100=closed), inverted vs HA convention
        return True

    @property
    def is_slider(self) -> bool:
        return True

    @property
    def has_tilt(self) -> bool:
        return False

    @property
    def open_command(self) -> str:
        return "open"

    @property
    def close_command(self) -> str:
        return "close"

    @property
    def is_calibrated(self) -> None:
        return None

    def stop_command(self, has_stop: bool) -> str:
        return "stop"

    def read_has_stop(self, alias: str, raw_value: Any, product: "Box") -> bool:
        return True


class Shutter(Gate):
    _control_type: Optional[ShutterBoxControlType]
    _is_calibrated: bool

    def __init__(self, control_type: int, is_calibrated: bool = False):
        super().__init__(control_type)
        self._is_calibrated = is_calibrated

    @property
    def min_position(self) -> int:
        return -1  # "unknown"

    @property
    def has_tilt(self) -> bool:
        return self._control_type in (
            ShutterBoxControlType.TILT_SHUTTER_90,
            ShutterBoxControlType.TILT_SHUTTER_180,
            ShutterBoxControlType.TILT_ONLY,
            ShutterBoxControlType.PERGOLA_ROOF_TILT_ONLY,
            ShutterBoxControlType.TILT_SHUTTER_WITHOUT_POSITIONING,
        )

    @property
    def tilt_only(self) -> bool:
        return self._control_type in (
            ShutterBoxControlType.TILT_ONLY,
            ShutterBoxControlType.PERGOLA_ROOF_TILT_ONLY,
        )

    @property
    def is_tilt_180(self) -> bool:
        return self._control_type == ShutterBoxControlType.TILT_SHUTTER_180

    @property
    def is_slider(self) -> bool:
        return (
            self._control_type
            not in (
                ShutterBoxControlType.NO_CALIBRATION,
                ShutterBoxControlType.CURTAIN,
                ShutterBoxControlType.TILT_ONLY,
                ShutterBoxControlType.PERGOLA_ROOF_TILT_ONLY,
                ShutterBoxControlType.TILT_SHUTTER_WITHOUT_POSITIONING,
            )
            and self.is_calibrated is True
        )

    @property
    def is_calibrated(self) -> bool:
        return self._is_calibrated

    def read_cover_type(
        self, alias: str, raw_value: Any, product: "Box"
    ) -> UnifiedCoverType:
        if self._control_type == ShutterBoxControlType.SEGMENTED_SHUTTER:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.NO_CALIBRATION:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.TILT_SHUTTER_90:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.WINDOW_OPENER:
            return UnifiedCoverType.WINDOW
        if self._control_type == ShutterBoxControlType.MATERIAL_SHUTTER:
            return UnifiedCoverType.SHADE
        if self._control_type == ShutterBoxControlType.AWNING:
            return UnifiedCoverType.AWNING
        if self._control_type == ShutterBoxControlType.SCREEN:
            return UnifiedCoverType.SHADE
        if self._control_type == ShutterBoxControlType.CURTAIN:
            return UnifiedCoverType.CURTAIN
        if self._control_type == ShutterBoxControlType.ROOF_SEGMENTED_SHUTTER:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.SERVOMOTOR:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.CURTAIN_ONE_SIDED:
            return UnifiedCoverType.CURTAIN
        if self._control_type == ShutterBoxControlType.CURTAIN_TWO_SIDED:
            return UnifiedCoverType.CURTAIN
        if self._control_type == ShutterBoxControlType.TILT_ONLY:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.PERGOLA_ROOF:
            return UnifiedCoverType.AWNING
        if self._control_type == ShutterBoxControlType.PERGOLA_ROOF_TILT_ONLY:
            return UnifiedCoverType.AWNING
        if self._control_type == ShutterBoxControlType.TILT_SHUTTER_180:
            return UnifiedCoverType.SHUTTER
        if self._control_type == ShutterBoxControlType.TILT_SHUTTER_WITHOUT_POSITIONING:
            return UnifiedCoverType.SHUTTER


class GateBox(Gate):
    _control_type: Optional[GateBoxControlType]

    @property
    def is_position_inverted(self) -> bool:
        # gateBox: (0=closed, 100=open), matches HA convention
        return False

    @property
    def is_slider(self) -> bool:
        return False

    @property
    def open_command(self) -> str:
        return "primary"

    @property
    def close_command(self) -> str:
        return "primary"

    def read_state(self, alias: str, raw_value: Any, product: "Box") -> int:
        current = raw_value("position")
        desired = raw_value("desired")

        # gate with gateBox visualized:
        #  (0) [   <#####] (100)
        if current == -1:
            return None

        if desired < current:
            return 0  # closing

        if desired > current:
            return 1  # opening

        if current == 0:  # closed
            return 3  # closed (lower/left limit)

        if current == 100:  # opened
            return 4  # open (upper/right limit)

        return 2  # manually stopped

    def stop_command(self, has_stop: bool) -> str:
        if not has_stop:
            raise MisconfiguredDevice("second button not configured as 'stop'")
        return "secondary"

    def read_has_stop(self, alias: str, raw_value: Any, product: "Box") -> bool:
        if product.last_data is None:
            return False

        button_type = raw_value("extraButtonType")
        if button_type is None:
            return False

        return button_type == 1


class GateBoxB(GateBox):
    def read_state(self, alias: str, raw_value: Any, product: "Box") -> int:
        current = raw_value("position")

        # gate with gateBox visualized:
        #  (0) [   <#####] (100)
        if current == -1:
            return None

        elif 0 < current < 100:
            return 2  # manually stopped
        elif current == 0:  # closed
            return 3  # closed (lower/left limit)

        return 4  # open (upper/right limit)

    def read_desired(self, alias: str, raw_value: Any, product: "Box") -> Optional[int]:
        return raw_value("position")

    def read_has_stop(self, alias: str, raw_value: Any, product: "Box") -> bool:
        # note: if control type is unknown we assume it is not open/close
        #       and has the stop feature via secondary button command.
        return self._control_type != GateBoxControlType.OPEN_CLOSE

    def read_cover_type(
        self, alias: str, raw_value: Any, product: "Box"
    ) -> Optional[UnifiedCoverType]:
        if (gate_type := raw_value("gate_type")) is None:
            return

        if gate_type == GateBoxGateType.GARAGE_DOOR:
            return UnifiedCoverType.GARAGE
        if gate_type == GateBoxGateType.SLIDING_DOOR:
            return UnifiedCoverType.GATE
        return UnifiedCoverType.DOOR

    @property
    def close_command(self) -> str:
        if self._control_type == GateBoxControlType.OPEN_CLOSE:
            return "secondary"

        return super().close_command


GateT = TypeVar("GateT", bound=Gate)


# TODO: handle tilt
class Cover(Feature):
    _desired: Optional[int]
    _state: Optional[BleboxCoverState]
    _has_stop: Optional[bool]
    _cover_type: Optional[UnifiedCoverType]

    def __init__(
        self,
        product: "Box",
        alias: str,
        methods: dict,
        dev_class: str,
        subclass: Type[GateT],
        extended_state: dict,
    ) -> None:
        control_type = None
        is_calibrated: bool = False
        if extended_state and issubclass(subclass, Shutter):
            control_type = extended_state.get("shutter", {}).get("controlType", None)
            is_calibrated = bool(
                extended_state.get("shutter", {})
                .get("calibrationParameters", {})
                .get("isCalibrated", False)
            )

        elif extended_state and issubclass(subclass, GateBoxB):
            control_type = extended_state.get("gate", {}).get("openCloseMode", None)

        self._device_class = dev_class
        self._attributes: GateT = subclass(control_type, is_calibrated)
        self._tilt_current = None
        super().__init__(product, alias, methods)

    @classmethod
    def many_from_config(
        cls, product, box_type_config, extended_state
    ) -> list["Feature"]:
        return [cls(product, *args, extended_state) for args in box_type_config]

    @property
    def current(self) -> Any:
        return self._desired

    @property
    def state(self) -> Any:
        return self._state

    @property
    def tilt_current(self):
        return self._tilt_current

    @property
    def is_calibrated(self) -> Optional[bool]:
        return self._attributes.is_calibrated

    @property
    def is_slider(self) -> Any:
        return self._attributes.is_slider

    @property
    def has_tilt(self) -> bool:
        return self._attributes.has_tilt

    @property
    def is_position_inverted(self) -> bool:
        return self._attributes.is_position_inverted

    @property
    def has_stop(self) -> bool:
        return self._has_stop

    @property
    def tilt_only(self) -> bool:
        return (
            self._attributes.tilt_only
            if hasattr(self._attributes, "tilt_only")
            else False
        )

    @property
    def is_tilt_180(self) -> bool:
        return (
            self._attributes.is_tilt_180
            if hasattr(self._attributes, "is_tilt_180")
            else False
        )

    @property
    def cover_type(self) -> Optional[UnifiedCoverType]:
        return self._cover_type

    async def async_open(self) -> None:
        await self.async_api_command(self._attributes.open_command)

    async def async_close(self) -> None:
        await self.async_api_command(self._attributes.close_command)

    async def async_stop(self) -> None:
        await self.async_api_command(self._attributes.stop_command(self._has_stop))

    async def async_set_position(self, value: Any) -> None:
        if not self.is_slider:
            raise NotImplementedError

        await self.async_api_command("position", value)

    async def async_set_tilt_position(self, value: Any) -> None:
        if self.has_tilt:
            await self.async_api_command("tilt", value)
        else:
            raise NotImplementedError

    async def async_close_tilt(self, **kwargs: Any) -> None:
        await self.async_api_command("tilt", 100)

    async def async_open_tilt(self, **kwargs: Any) -> None:
        await self.async_api_command("tilt", 0)

    def _read_cover_type(self) -> Optional[UnifiedCoverType]:
        product = self._product
        if not product.last_data:
            return None

        alias = self._alias
        return self._attributes.read_cover_type(alias, self.raw_value, self._product)

    def _read_desired(self) -> Any:
        product = self._product
        if not product.last_data:
            return None

        alias = self._alias
        return self._attributes.read_desired(alias, self.raw_value, self._product)

    # TODO: refactor
    def _read_state(self) -> Any:
        product = self._product
        if not product.last_data:
            return None

        alias = self._alias
        return self._attributes.read_state(alias, self.raw_value, self._product)

    def _read_tilt(self) -> Any:
        product = self._product
        if not product.last_data:
            return None

        alias = self._alias
        return self._attributes.read_tilt(alias, self.raw_value, self._product)

    def _read_has_stop(self) -> bool:
        return self._attributes.read_has_stop(
            self._alias, self.raw_value, self._product
        )

    def after_update(self) -> None:
        self._desired = self._read_desired()
        self._state = self._read_state()
        self._has_stop = self._read_has_stop()
        self._cover_type = self._read_cover_type()

        if self._attributes.has_tilt:
            self._tilt_current = self._read_tilt()
